#define WIN32_LEAN_AND_MEAN        // trim the excess fat from Windows

/*******************************************************************
*    Program: Chapter 17 DirectX Audio - DirectMusic Example 2
*    Author: Kevin Hawkins
*    Description: Illustrates how to use DirectX Audio for 3D
*                 audio effects. An object making a sound
*                 moves close to and away from the camera.
********************************************************************/

#define INITGUID                   // we use GUID's with DMusic

////// Includes
#include <windows.h>               // standard Windows app include
#include <dmusicc.h>               // DirectMusic includes
#include <dmusici.h>
#include <d3d8types.h>             // for D3DVECTOR
#include <cguid.h>                 // for GUID_NULL
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <gl/gl.h>                 // standard OpenGL include
#include <gl/glu.h>                // OpenGL utilties

////// Global Variables
HDC g_HDC;                         // global device context
bool fullScreen = false;           // true = fullscreen; false = windowed
bool keyPressed[256];              // holds true for keys that are pressed 

float angle = 0.0f;
unsigned int listBase;             // display list base
GLYPHMETRICSFLOAT gmf[256];        // holds orientation and placement
                                   // info for display lists

////// DirectMusic variables
IDirectMusicLoader8 *dmusicLoader = NULL;              // the loader
IDirectMusicPerformance8 *dmusicPerformance = NULL;    // the performance
IDirectMusicSegment8 *dmusicSegment = NULL;            // the segment
IDirectMusicAudioPath *dmusic3DAudioPath = NULL;       // the audiopath
IDirectSound3DBuffer *ds3DBuffer = NULL;               // 3d buffer
IDirectSound3DListener *ds3DListener = NULL;           // 3d listener

DS3DBUFFER dsBufferParams;                             // 3d buffer properties
DS3DLISTENER dsListenerParams;                         // 3d listener properties

/*******************************************************
*    DirectMusic Interfaces
*******************************************************/

// InitDirectXAudio()
// desc: initializes the DirectX Audio component for playback
bool InitDirectXAudio(HWND hwnd)
{
     char pathStr[MAX_PATH];       // path for audio file
     WCHAR wcharStr[MAX_PATH];

     // create the loader object
     if (FAILED(CoCreateInstance(CLSID_DirectMusicLoader, NULL, CLSCTX_INPROC,
                                 IID_IDirectMusicLoader8, (void**)&dmusicLoader)))
     {
          MessageBox(hwnd, "Unable to create the IDirectMusicLoader8 object!\nPress OK to exit",
                     "ERROR!", MB_OK);
          return false;
     }

     // create the performance object
     if (FAILED(CoCreateInstance(CLSID_DirectMusicPerformance, NULL, CLSCTX_INPROC,
                                 IID_IDirectMusicPerformance8, (void**)&dmusicPerformance)))
     {
          MessageBox(hwnd, "Unable to create the IDirectMusicPerformance8 object!\nPress OK to exit",
                     "ERROR!", MB_OK);
          return false;
     }

     // intialize DirectMusic and DirectSound
     if (FAILED(dmusicPerformance->InitAudio(NULL, NULL, NULL, DMUS_APATH_DYNAMIC_STEREO, 64,
                                  DMUS_AUDIOF_ALL, NULL)))
     {
          MessageBox(hwnd, "Unable to initialize audio! Press OK to exit", "ERROR!", MB_OK);
          return false;
     }

     // create a standard 3D audiopath
     if (FAILED(dmusicPerformance->CreateStandardAudioPath(DMUS_APATH_DYNAMIC_3D,
                                                           64, TRUE, &dmusic3DAudioPath)))
     {
          MessageBox(hwnd, "Unable to create standard 3D audiopath! Press OK to exit",
                     "ERROR!", MB_OK);
          return false;
     }

     // get the 3D buffer in the audiopath
     if (FAILED(dmusic3DAudioPath->GetObjectInPath(0, DMUS_PATH_BUFFER, 0, GUID_NULL, 0,
                                                   IID_IDirectSound3DBuffer,
                                                   (void**)&ds3DBuffer)))
     {
          MessageBox(hwnd, "Unable to retrieve 3D buffer from audiopath! Press OK to exit",
                     "ERROR!", MB_OK);
          return false;
     }

     // get the 3D buffer parameters
     dsBufferParams.dwSize = sizeof(DS3DBUFFER);
     ds3DBuffer->GetAllParameters(&dsBufferParams);

     // set the new 3D buffer parameters
     dsBufferParams.dwMode = DS3DMODE_HEADRELATIVE;    // relative to the listener
     ds3DBuffer->SetAllParameters(&dsBufferParams, DS3D_IMMEDIATE);

     // retrieve the listener from the audiopath
     if (FAILED(dmusic3DAudioPath->GetObjectInPath(0, DMUS_PATH_PRIMARY_BUFFER, 0, GUID_NULL, 0, 
                                                  IID_IDirectSound3DListener,
                                                  (void**)&ds3DListener)))
     {
          MessageBox(hwnd, "Unable to retrieve the listener! Press OK to exit",
                     "ERROR!", MB_OK);
          return false;
     }

     // get the listener parameters
     dsListenerParams.dwSize = sizeof(DS3DLISTENER);
     ds3DListener->GetAllParameters(&dsListenerParams);

     // set position of listener
     dsListenerParams.vPosition.x = 0.0f;
     dsListenerParams.vPosition.y = 0.0f;
     dsListenerParams.vPosition.z = 0.0f;
     ds3DListener->SetAllParameters(&dsListenerParams, DS3D_IMMEDIATE);

     // retrieve the current directory
     GetCurrentDirectory(MAX_PATH, pathStr);

     // convert to unicode string
     MultiByteToWideChar(CP_ACP, 0, pathStr, -1, wcharStr, MAX_PATH);

     // set the search directory
     dmusicLoader->SetSearchDirectory(GUID_DirectMusicAllTypes, wcharStr, FALSE);

     return true;
}

// LoadSegment()
// desc: load a segment from a file
bool LoadSegment(HWND hwnd, char *filename)
{
     WCHAR wcharStr[MAX_PATH];

     // convert filename to unicode string
     MultiByteToWideChar(CP_ACP, 0, filename, -1, wcharStr, MAX_PATH);

     // load the segment from file
     if (FAILED(dmusicLoader->LoadObjectFromFile(CLSID_DirectMusicSegment,
                                                 IID_IDirectMusicSegment8,
                                                 wcharStr,
                                                 (void**)&dmusicSegment)))
     {
          MessageBox(hwnd, "Audio file not found! Press OK to exit",
                     "ERROR!", MB_OK);

          return false;
     }

     // set the number of repeats for the segment to infinite
     dmusicSegment->SetRepeats(DMUS_SEG_REPEAT_INFINITE);

     // download the segment's instruments to the audiopath
     dmusicSegment->Download(dmusic3DAudioPath);

     return true;
}

// PlaySegment()
// desc: start playing a segment
void PlaySegment(IDirectMusicPerformance8* dmPerf, IDirectMusicSegment8* dmSeg)
{
     // play the segment on the next beat
     dmPerf->PlaySegmentEx(dmSeg, NULL, NULL, DMUS_SEGF_DEFAULT, 0,
                           NULL, NULL, dmusic3DAudioPath);
}

// StopSegment()
// desc: stop a segment from playing
void StopSegment(IDirectMusicPerformance8* dmPerf, IDirectMusicSegment8* dmSeg)
{
     // stop the dmSeg from playing
     dmPerf->StopEx(dmSeg, 0, 0);
}

// CloseDown()
// desc: shutdown music performance
void CloseDown(IDirectMusicPerformance8* dmPerf)
{
     // stop the music
     dmPerf->Stop(NULL, NULL, 0, 0);

     // close down DirectMusic
     dmPerf->CloseDown();
}

// Set3DSoundParams()
// desc: sets the 3d buffer parameters
void Set3DSoundParams(float doppler, float rolloff, float minDist, float maxDist)
{
     // set doppler and rolloff parameters
     dsListenerParams.flDopplerFactor = doppler;
     dsListenerParams.flRolloffFactor = rolloff;

     if (ds3DListener)
          ds3DListener->SetAllParameters(&dsListenerParams, DS3D_IMMEDIATE);

     // set minimum and maximum distances
     dsBufferParams.flMinDistance = minDist;
     dsBufferParams.flMaxDistance = maxDist;

     if (ds3DBuffer)
          ds3DBuffer->SetAllParameters(&dsBufferParams, DS3D_IMMEDIATE);
}

// Set3DSoundPos()
// desc: updates position of sound source (accepts OpenGL coordinates)
void Set3DSoundPos(IDirectSound3DBuffer* dsBuff, float x, float y, float z)
{
     // we use -z because DirectX and OpenGL z-axes are flipped
     if (dsBuff != NULL)
     {
          dsBuff->SetPosition(x, y, -z, DS3D_IMMEDIATE);
     }
}

/******************************************************
*    OpenGL Interfaces
******************************************************/

// CreateOutlineFont()
// desc: creates the outline font using the CreateFont() function
unsigned int CreateOutlineFont(char *fontName, int fontSize, float depth)
{
     HFONT hFont;         // windows font
     unsigned int base;

     base = glGenLists(256);      // create storage for 96 characters

     if (stricmp(fontName, "symbol") == 0)
     {
          hFont = CreateFont(fontSize, 0, 0, 0, FW_BOLD, FALSE, FALSE, FALSE, SYMBOL_CHARSET, 
                                   OUT_TT_PRECIS, CLIP_DEFAULT_PRECIS, ANTIALIASED_QUALITY,
                                   FF_DONTCARE | DEFAULT_PITCH, fontName);
     }
     else
     {
           hFont = CreateFont(fontSize, 0, 0, 0, FW_BOLD, FALSE, FALSE, FALSE, ANSI_CHARSET, 
                                   OUT_TT_PRECIS, CLIP_DEFAULT_PRECIS, ANTIALIASED_QUALITY,
                                   FF_DONTCARE | DEFAULT_PITCH, fontName);
     }

     if (!hFont)
          return 0;

     SelectObject(g_HDC, hFont);
     wglUseFontOutlines(g_HDC, 0, 255, base, 0.0f, depth, WGL_FONT_POLYGONS, gmf);

     return base;
}

// ClearFont()
// desc: deletes the display list for the font
void ClearFont(unsigned int base)
{
     glDeleteLists(base, 256);
}

// PrintString()
// desc: displays the text in str from the font indicated by base
void PrintString(unsigned int base, char *str)
{
     float length = 0;

     if ((str == NULL))
          return;

     // center the text
     for (unsigned int loop=0;loop<(strlen(str));loop++)    // find length of text
     {
          length+=gmf[str[loop]].gmfCellIncX;                  // increase length by character's width
     }
     glTranslatef(-length/2,0.0f,0.0f);                  // translate to center text

     // draw the text
     glPushAttrib(GL_LIST_BIT);
          glListBase(base);
          glCallLists(strlen(str), GL_UNSIGNED_BYTE, str);
     glPopAttrib();
}

// CleanUp()
// desc: application cleanup
void CleanUp()
{
     ClearFont(listBase);

     dmusic3DAudioPath->Release();
     dmusicLoader->Release();
     dmusicPerformance->Release();
     dmusicSegment->Release();
}

// Initialize
// desc: initializes OpenGL
void Initialize()
{
     glClearColor(0.0f, 0.0f, 0.0f, 0.0f);        // clear to black

     glShadeModel(GL_SMOOTH);                     // use smooth shading
     glEnable(GL_DEPTH_TEST);                     // hidden surface removal
     glEnable(GL_LIGHT0);                         // enable light0
     glEnable(GL_LIGHTING);                       // enable lighting
     glEnable(GL_COLOR_MATERIAL);                 // enable color for material

     listBase = CreateOutlineFont("Arial", 10, 0.25f);  // load 10pt Arial font
}

// Render
// desc: handles drawing of scene
void Render()
{
     static float zpos = 0.0f;     // the position along the z-axis
     static bool zDir = false;     // false = negative, true = positive

     // move text forward and backward
     if (zDir)
          zpos += 0.08f;
     else
          zpos -= 0.08f;

     if (zpos > 30.0f)
          zDir = false;
     if (zpos < -30.0f)
          zDir = true;

     // clear screen and depth buffer
     glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);         
     glLoadIdentity();

     glPushMatrix();
          glColor3f(0.5f, 0.5f, 0.5f);
          glTranslatef(0.0f, 3.0f, -10.0f);
          PrintString(listBase, "Hear me click?");
     glPopMatrix();

     // move the text along the z-axis
     glTranslatef(0.0f, 0.0f, zpos);

     // update sound source position to reflect text
     Set3DSoundPos(ds3DBuffer, 0.0, 0.0, zpos);

     // set color to blue-ish color
     glColor3f(0.3f, 0.4f, 0.8f);

     // display the text
     glPushMatrix();
          glTranslatef(0.0f, -1.0f, 0.0f);
          if (zDir)
               PrintString(listBase, "Coming at you!");
          else
               PrintString(listBase, "Moving away from you!");
     glPopMatrix();

     // yellow-green color
     glColor3f(0.6f, 0.8f, 0.5f);

     // display text
     glPushMatrix();
          glTranslatef(0.0f, -3.0f, 0.0f);
          PrintString(listBase, "ESC - Quit");
     glPopMatrix();

     SwapBuffers(g_HDC);           // bring backbuffer to foreground
}

// function to set the pixel format for the device context
void SetupPixelFormat(HDC hDC)
{
     int nPixelFormat;                       // our pixel format index

     static PIXELFORMATDESCRIPTOR pfd = {
          sizeof(PIXELFORMATDESCRIPTOR),     // size of structure
          1,                                 // default version
          PFD_DRAW_TO_WINDOW |               // window drawing support
          PFD_SUPPORT_OPENGL |               // OpenGL support
          PFD_DOUBLEBUFFER,                  // double buffering support
          PFD_TYPE_RGBA,                     // RGBA color mode
          32,                                // 32 bit color mode
          0, 0, 0, 0, 0, 0,                  // ignore color bits, non-palettized mode
          0,                                 // no alpha buffer
          0,                                 // ignore shift bit
          0,                                 // no accumulation buffer
          0, 0, 0, 0,                        // ignore accumulation bits
          16,                                // 16 bit z-buffer size
          0,                                 // no stencil buffer
          0,                                 // no auxiliary buffer
          PFD_MAIN_PLANE,                    // main drawing plane
          0,                                 // reserved
          0, 0, 0 };                         // layer masks ignored

     nPixelFormat = ChoosePixelFormat(hDC, &pfd); // choose best matching pixel format

     SetPixelFormat(hDC, nPixelFormat, &pfd);     // set pixel format to device context
}

// the Windows Procedure event handler
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
     static HGLRC hRC;                       // rendering context
     static HDC hDC;                         // device context
     int width, height;                      // window width and height

     switch(message)
     {
          case WM_CREATE:                    // window is being created

               hDC = GetDC(hwnd);            // get current window's device context
               g_HDC = hDC;
               SetupPixelFormat(hDC);        // call our pixel format setup function

               // create rendering context and make it current
               hRC = wglCreateContext(hDC);
               wglMakeCurrent(hDC, hRC);

               return 0;
               break;

          case WM_CLOSE:                     // windows is closing

               // deselect rendering context and delete it
               wglMakeCurrent(hDC, NULL);
               wglDeleteContext(hRC);

               // send WM_QUIT to message queue
               PostQuitMessage(0);

               return 0;
               break;

          case WM_SIZE:
               height = HIWORD(lParam);      // retrieve width and height
               width = LOWORD(lParam);

               if (height==0)                // don't want a divide by zero
               {
                    height=1;                     
               }

               glViewport(0, 0, width, height);   // reset the viewport to new dimensions
               glMatrixMode(GL_PROJECTION);       // set projection matrix current matrix
               glLoadIdentity();                  // reset projection matrix

               // calculate aspect ratio of window
               gluPerspective(54.0f,(GLfloat)width/(GLfloat)height,1.0f,1000.0f);

               glMatrixMode(GL_MODELVIEW);        // set modelview matrix
               glLoadIdentity();                  // reset modelview matrix

               return 0;
               break;

          case WM_KEYDOWN:                        // is a key pressed?
               keyPressed[wParam] = true;
               return 0;
               break;

          case WM_KEYUP:
               keyPressed[wParam] = false;
               return 0;
               break;

          default:
               break;
     }

     return (DefWindowProc(hwnd, message, wParam, lParam));
}

// the main windows entry point
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
     WNDCLASSEX windowClass;  // window class
     HWND    hwnd;            // window handle
     MSG          msg;        // message
     bool    done;            // flag saying when our app is complete
     DWORD        dwExStyle;  // Window Extended Style
     DWORD        dwStyle;    // Window Style
     RECT    windowRect;

     // screen resolution
     int width = 800;
     int height = 600;
     int bits = 16;

     //fullScreen = TRUE;

     windowRect.left=(long)0;                          // Set Left Value To 0
     windowRect.right=(long)width;                     // Set Right Value To Requested Width
     windowRect.top=(long)0;                           // Set Top Value To 0
     windowRect.bottom=(long)height;                   // Set Bottom Value To Requested Height

     // fill out the window class structure
     windowClass.cbSize            = sizeof(WNDCLASSEX);
     windowClass.style             = CS_HREDRAW | CS_VREDRAW;
     windowClass.lpfnWndProc       = WndProc;
     windowClass.cbClsExtra        = 0;
     windowClass.cbWndExtra        = 0;
     windowClass.hInstance         = hInstance;
     windowClass.hIcon             = LoadIcon(NULL, IDI_APPLICATION); // default icon
     windowClass.hCursor           = LoadCursor(NULL, IDC_ARROW);     // default arrow
     windowClass.hbrBackground     = NULL;                            // don't need background
     windowClass.lpszMenuName = NULL;                                 // no menu
     windowClass.lpszClassName     = "MyClass";
     windowClass.hIconSm           = LoadIcon(NULL, IDI_WINLOGO);     // windows logo small icon

     // register the windows class
     if (!RegisterClassEx(&windowClass))
          return 0;

     if (fullScreen)                                        // fullscreen?
     {
          DEVMODE dmScreenSettings;                         // device mode
          memset(&dmScreenSettings,0,sizeof(dmScreenSettings));
          dmScreenSettings.dmSize = sizeof(dmScreenSettings);    
          dmScreenSettings.dmPelsWidth = width;        // screen width
          dmScreenSettings.dmPelsHeight = height;      // screen height
          dmScreenSettings.dmBitsPerPel = bits;        // bits per pixel
          dmScreenSettings.dmFields=DM_BITSPERPEL|DM_PELSWIDTH|DM_PELSHEIGHT;

          // 
          if (ChangeDisplaySettings(&dmScreenSettings, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
          {
               // setting display mode failed, switch to windowed
               MessageBox(NULL, "Display mode failed", NULL, MB_OK);
               fullScreen=FALSE;   
          }
     }

     if (fullScreen)                                   // Are We Still In Fullscreen Mode?
     {
          dwExStyle=WS_EX_APPWINDOW;                   // Window Extended Style
          dwStyle=WS_POPUP;                            // Windows Style
          ShowCursor(FALSE);                           // Hide Mouse Pointer
     }
     else
     {
          dwExStyle=WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;     // Window Extended Style
          dwStyle=WS_OVERLAPPEDWINDOW;                      // Windows Style
     }

     AdjustWindowRectEx(&windowRect, dwStyle, FALSE, dwExStyle);      // Adjust Window To True Requested Size

     // class registered, so now create our window
     hwnd = CreateWindowEx(NULL,                                           // extended style
                                "MyClass",                                 // class name
                                "DirectX Audio Example 2 - 3D Effects",// app name
                                dwStyle | WS_CLIPCHILDREN |
                                WS_CLIPSIBLINGS,
                                0, 0,                                 // x,y coordinate
                                windowRect.right - windowRect.left,
                                windowRect.bottom - windowRect.top,   // width, height
                                NULL,                                 // handle to parent
                                NULL,                                 // handle to menu
                                hInstance,                            // application instance
                                NULL);                                // no extra params

     // check if window creation failed (hwnd would equal NULL)
     if (!hwnd)
          return 0;

     // initialize COM
     if (FAILED(CoInitialize(NULL)))
          return 0;                     // if unable to initialize COM, exit

     // initialize DirectX Audio
     if (!InitDirectXAudio(hwnd))
          return 0;

     // load the segment
     if (!LoadSegment(hwnd, "start.wav"))
          return 0;

     // set the 3D sound parameters
     Set3DSoundParams(0.0, 0.1f, 1.0f, 100.0f);

     // start the segment (infinitely loops)
     PlaySegment(dmusicPerformance, dmusicSegment);

     ShowWindow(hwnd, SW_SHOW);              // display the window
     UpdateWindow(hwnd);                     // update the window

     done = false;                           // intialize the loop condition variable
     Initialize();                           // initialize OpenGL

     // main message loop
     while (!done)
     {
          PeekMessage(&msg, hwnd, NULL, NULL, PM_REMOVE);

          if (msg.message == WM_QUIT)        // do we receive a WM_QUIT message?
          {
               done = true;                  // if so, time to quit the application
          }
          else
          {
               // exit the app
               if (keyPressed[VK_ESCAPE])
                    done = true;
               else
               {
                    Render();                // render a frame

                    TranslateMessage(&msg);  // translate and dispatch to event queue
                    DispatchMessage(&msg);
               }
          }
     }

     // close down the performance
     CloseDown(dmusicPerformance);

     // clean up application
     CleanUp();

     // close down COM
     CoUninitialize();

     if (fullScreen)
     {
          ChangeDisplaySettings(NULL,0);          // If So Switch Back To The Desktop
          ShowCursor(TRUE);                       // Show Mouse Pointer
     }

     return msg.wParam;
}
